class SettingsContainer:
    __writable__ = False

    dryrun = False

    def __enter__(self):
        self.__writable__ = True

    def __exit__(self, *args, **kwargs):
        self.__writable__ = False

    def __setattr__(self, key, value):
        if key != '__writable__':
            if not hasattr(self, key):
                raise AttributeError(f'Unknown settings key: {key}')
            if not self.__writable__:
                raise RuntimeError(
                    f'Settings not writable unless inside context manager'
                )
            if not isinstance(value, type(getattr(self, key))):
                raise TypeError(
                    f'Settings key {key} currently has type '
                    f'{type(getattr(self, key))} but you passed in '
                    f'{value} which is {type(value)}'
                )
        super().__setattr__(key, value)


settings = SettingsContainer()



import pytest

from ruv_dl.runtime import settings


def test_cannot_set_unless_context_manager():
    with pytest.raises(RuntimeError):
        settings.dryrun = True
    assert settings.dryrun is False


def test_sets_dryrun_using_context_manager():
    with settings:
        settings.dryrun = True
    assert settings.dryrun is True
    # Reset again
    with settings:
        settings.dryrun = False
    assert settings.dryrun is False


def test_must_use_same_type_when_setting_new_value():
    with settings:
        with pytest.raises(TypeError):
            settings.dryrun = 'not a boolean'
    assert settings.dryrun is False


def test_cannot_set_unknown_setting():
    with settings:
        with pytest.raises(AttributeError):
            settings.not_a_known_key = 'something'



#!/usr/bin/env python
import datetime
import logging

from urllib.parse import parse_qs, urlparse
import requests

from ruv_dl.cache import DiskCache
from ruv_dl.data import Entry
from ruv_dl.constants import (
    DATETIME_FORMAT, DATE_FORMAT, URL_TEMPLATE, DATE_PART_LENGTH
)

logger = logging.getLogger(__name__)


class Crawler:
    def __init__(
        self, program, iteration_count, days_between_episodes
    ):
        self.program = program
        self.itercount = iteration_count
        self.days_between_episodes = days_between_episodes
        self.prefer_open = True
        self.cache = DiskCache(program['id'])
        logger.debug(
            '\n'.join([
                'Initializing crawler with:',
                f'Iteration count: {self.itercount}',
                f'Days between episodes: {self.days_between_episodes}',
            ])
        )

    def get_entry(self, date, fn, episode=None):
        cache_key = f'{date.strftime(DATE_FORMAT)}-{fn}'
        if (
            not self.cache.has(cache_key)
        ):
            r = requests.head(
                URL_TEMPLATE.format(
                    date=date.strftime(DATE_FORMAT),
                    fn=fn,
                    openclose='opid' if self.prefer_open else 'lokad',
                )
            )
            logger.info(
                'Checking %s - %s - %s (is_open: %s)' % (
                    date.strftime(DATE_FORMAT),
                    fn,
                    r.ok,
                    self.prefer_open,
                )
            )
            if r.ok:
                self.cache.set(
                    cache_key,
                    {
                        'success': True,
                        'url': r.url,
                        'etag': r.headers['ETag'],
                        'checked_at': datetime.datetime.now().strftime(
                            DATETIME_FORMAT,
                        ),
                    }
                )
            else:
                self.cache.set(
                    cache_key,
                    {
                        'success': False,
                        'status_code': r.status_code,
                        'checked_at': datetime.datetime.now().strftime(
                            DATETIME_FORMAT,
                        )
                    }
                )
        info = self.cache.get(cache_key)
        checked_at = datetime.datetime.strptime(
            info['checked_at'],
            DATETIME_FORMAT,
        )
        if info['success']:
            return Entry(
                fn=fn,
                url=info['url'],
                date=date,
                etag=info['etag'],
                episode=episode,
            )
        elif (
            # Don't remove unless we last checked before the show was aired
            checked_at <= (date + datetime.timedelta(1)) and
            # And we haven't checked this link for over 1 hours
            abs(
                (checked_at - datetime.datetime.now()).total_seconds() / 3600
            ) > 1 and
            # And the show should have been aired
            date <= (datetime.datetime.now() + datetime.timedelta(1))
        ):
            self.cache.remove(cache_key)
            return self.get_entry(date, fn)

    def get_new_fn(self, fn, direction):
        known_delimeters = 'ATS'
        for delimiter in known_delimeters:
            try:
                fn_id, something = fn.split(delimiter)
            except ValueError:
                continue
            new_id = str(int(fn_id) + direction)
            while len(new_id) < len(fn_id):
                new_id = f'0{new_id}'
            return f'{new_id}{delimiter}{something}'
        else:
            raise RuntimeError(
                f'No known delimiters [{known_delimeters}] found in {fn}'
            )

    def crawl(self, date, fn, direction=1):
        new_fn = self.get_new_fn(fn, direction)
        for i in range(self.itercount):
            # Search for maximum 2 weeks back in time
            date_to_check = (
                date +
                datetime.timedelta(
                    days=i * direction * self.days_between_episodes
                )
            )
            entry = self.get_entry(
                date_to_check,
                new_fn,
            )
            if entry:
                yield entry
                yield from self.crawl(date_to_check, new_fn, direction)
                break

    def search_for_episodes(self):
        files = set()
        episodes = self.program['episodes']
        if not episodes:
            raise RuntimeError('No episodes found...')
        for episode in episodes:
            self.prefer_open = 'opid' in episode['file']
            manifest_url = episode['file']
            parts = urlparse(manifest_url)
            query = parse_qs(parts.query)
            wanted_stream = query['streams'][0].split(',')[0]
            # Dates are the first part, '%Y/%m/%d'
            datestr = wanted_stream[:DATE_PART_LENGTH]
            try:
                date = datetime.datetime.strptime(
                    datestr,
                    DATE_FORMAT,
                )
            except ValueError:
                logger.info(
                    f'Could not parse date {datestr} from {wanted_stream}'
                )
                continue
            fn = wanted_stream.split('/')[-1].split('.')[0]
            first_entry = self.get_entry(date, fn, episode=episode)
            if not first_entry:
                raise RuntimeError('Could not get url for first episode...?')
            files.add(first_entry)
            logger.debug('Searching backwards in time...')
            for entry in self.crawl(date, fn, direction=-1):
                files.add(entry)
            logger.debug('Searching forward in time...')
            for entry in self.crawl(date, fn, direction=1):
                files.add(entry)
        self.cache.write()
        return files



import logging
import tempfile
import os
import copy
import shutil

from ruv_dl.programs import ProgramFetcher
from ruv_dl.data import Entry, Episode

logger = logging.getLogger(__name__)

'''
    TODO: Split migrations to seperate files when/if we need more.
          Create some kind of autodiscovery and autopopulation of migrations.
          Also automatically bump version in program_info?
'''


def move_old_locations(destination, dryrun=False):
    fetcher = ProgramFetcher(None, None, destination)
    for program_info in fetcher.get_all_program_infos():
        if program_info.version >= 1:
            logger.info('Skipping %s', program_info)
            continue
        program = program_info.program
        logger.info('Targeting %s', program['title'])
        seasons = program_info.seasons
        files_to_move = []
        for season, entries in seasons.items():
            season_folder = Entry.get_season_folder(
                destination, program, season
            )
            for i, target_entry in enumerate(entries.sorted()):
                if target_entry.episode.number is None:
                    raise RuntimeError(
                        'You need to attempt sync for this program once '
                        'before running this migration.'
                    )
                src_entry = copy.deepcopy(target_entry)
                src_entry.episode = Episode(None)
                src_entry.episode.number = i + 1

                src_dest = os.path.join(
                    season_folder,
                    src_entry.get_target_basename(program, season),
                )

                target_dest = os.path.join(
                    season_folder,
                    target_entry.get_target_basename(program, season),
                )

                if src_dest != target_dest and os.path.isfile(src_dest):
                    files_to_move.append((src_dest, target_dest))
        while files_to_move:
            i = 0
            src_dest, target_dest = files_to_move.pop(0)
            if dryrun:
                logger.info('Would move %s to %s', src_dest, target_dest)
            elif os.path.isfile(target_dest):
                temp_path = next(tempfile._get_candidate_names())
                temp_path = os.path.join(
                    os.path.dirname(target_dest),
                    temp_path,
                )
                logger.info(
                    '%s exists. Moving temporarily to %s and postponing',
                    target_dest, temp_path,
                )
                shutil.move(src_dest, temp_path)
                files_to_move.append((temp_path, target_dest))
            else:
                os.makedirs(os.path.dirname(target_dest), exist_ok=True)
                logger.info('Moving %s to %s', src_dest, target_dest)
                shutil.move(src_dest, target_dest)

        if not dryrun:
            program_info.version = 1
            program_info.write()


MIGRATIONS = {
    1: (move_old_locations, )
}



import datetime

from ruv_dl.data import Entry, EntrySet, Episode


def test_entry_from_dict():
    data = {
        'fn': 'some_fn',
        'url': 'some_url',
        'date': '2017/06/14',
        'etag': 'some_etag',
        'episode': {'id': 'some_episode'},
    }
    e = Entry.from_dict(data)
    assert e.fn == data['fn']
    assert e.url == data['url']
    assert e.date == datetime.datetime(2017, 6, 14)
    assert e.etag == data['etag']
    assert e.episode.data == data['episode']

    del data['episode']
    e = Entry.from_dict(data)
    assert e.episode.data == {}


def test_entry_to_dict():
    e = Entry(
        'some_fn', 'some_url', datetime.datetime(2017, 6, 14),
        'some_etag', {'id': 'some_episode'}
    )
    assert e.to_dict() == {
        'fn': 'some_fn',
        'url': 'some_url',
        'date': '2017/06/14',
        'etag': 'some_etag',
        'episode': {'id': 'some_episode'},
    }


def test_entry_equality():
    assert Entry('fdsa', 'fdsa', None, 'etag') == Entry(
        'asdf', 'asdf', datetime.datetime.min, 'etag'
    )
    assert Entry('fdsa', 'fdsa', None, 'etag') == Entry(
        'fdsa', 'fdsa', None, 'etag'
    )
    assert Entry('fdsa', 'fdsa', None, 'etag') != Entry(
        'fdsa', 'fdsa', None, 'not-same-etag'
    )


def test_choose_best_item():
    s = EntrySet()
    episode_real = {
        'id': 'some_id_from_ruv_api',
        'number': 1,
    }
    episode_generated = {
        'number': 2,
    }
    item1 = Entry('fn1', 'url1', datetime.datetime.min, 'etag1', None)
    item2 = Entry(
        'fn2', 'url2', datetime.datetime.min, 'etag2', episode_real
    )
    assert s._choose_best_item(item1, item2).episode.to_dict() == episode_real
    assert s._choose_best_item(item2, item1).episode.to_dict() == episode_real

    item1.episode = Episode(episode_generated)
    assert s._choose_best_item(item1, item2).episode.to_dict() == episode_real
    assert s._choose_best_item(item2, item1).episode.to_dict() == episode_real

    item2.episode = None
    assert s._choose_best_item(
        item1, item2).episode.to_dict() == episode_generated
    assert s._choose_best_item(
        item2, item1).episode.to_dict() == episode_generated



import pytest
import os
import datetime

from ruv_dl.data import EntrySet, Entry
from ruv_dl.programs import ProgramInfo
from ruv_dl.mover import Mover
from ruv_dl.exceptions import NamingSchemaError, ProgramInfoError


def create_program_info(location='/tv/Program', seasons=None):
    pi = ProgramInfo(location, initialize_empty=True)
    pi.program = {'id': 'some-id', 'title': 'Program'}
    pi.seasons = {
        key: EntrySet([
            Entry(
                '', '', datetime.datetime(2020, 1, 10), f'e{number}',
                episode={'number': number}
            )
            for number in value
        ])
        for key, value in (seasons or {}).items()
    }
    pi.write()
    return pi


def test_src_must_exist(fs):
    with pytest.raises(FileNotFoundError):
        Mover('/not/exists/', '/no/matter').move()


def test_cannot_move_to_destination_if_exists(fs):
    with open('/src', 'w') as f:
        f.write('contents')
    with open('/dst', 'w') as f:
        f.write('contents2')
    with pytest.raises(FileExistsError):
        Mover('/src', '/dst')


def test_src_and_dst_must_differ(fs):
    # This is essentially covered but I want to be explicit
    with open('/src', 'w') as f:
        f.write('contents')
    with pytest.raises(FileExistsError):
        Mover('/src', '/src')


def test_raises_if_no_program_info_found(fs):
    os.makedirs('/tv/Program/Source')
    os.chdir('/tv/Program')
    with pytest.raises(FileNotFoundError) as e:
        Mover('/tv/Program/Source', '/tv/Program/Destination')
    assert 'Could not find program info related to' in str(e.value)


def test_requires_absolute_path_to_move(fs):
    expected_error = (
        'Must pass in absolute path - "relative/path/here.mp4" is relative.'
    )
    with pytest.raises(OSError) as e:
        Mover('relative/path/here.mp4', '/absolute/path')
    assert str(e.value) == expected_error
    with pytest.raises(OSError) as e:
        Mover('/absolute/path', 'relative/path/here.mp4')
    assert str(e.value) == expected_error


def test_move_season_requires_correct_naming_schema(fs):
    os.makedirs('/tv/Program/Season')
    create_program_info()
    with pytest.raises(NamingSchemaError) as e:
        Mover('/tv/Program/Season', '/tv/Program/Season 2').move()
    expected_error = 'Season number not determined from /tv/Program/Season'
    assert str(e.value) == expected_error

    os.makedirs('/tv/Program/Season 1')
    with pytest.raises(NamingSchemaError) as e:
        Mover('/tv/Program/Season 1', '/tv/Program/Season something').move()
    expected_error = (
        'Season number not determined from /tv/Program/Season something'
    )
    assert str(e.value) == expected_error

    with pytest.raises(ProgramInfoError) as e:
        Mover('/tv/Program/Season 1', '/tv/Program/Season 2').move()
    expected_error = 'Source season number 1 not found in info file'
    assert str(e.value).startswith(expected_error)


def test_move_seasons_fails_if_any_episode_in_src_is_wrong_name_schema(fs):
    os.makedirs('/tv/Program/Season 1')
    create_program_info(seasons={1: [1]})
    os.chdir('/tv/Program')
    with open('/tv/Program/Season 1/Program - S02E01.mp4', 'w') as f:
        # Correct episode number, wrong season number
        f.write('S02E01')
    with pytest.raises(NamingSchemaError) as e:
        Mover('/tv/Program/Season 1', '/tv/Program/Season 2').move()
    expected_error = (
        'Season number in folder and filename differ in '
        '"/tv/Program/Season 1/Program - S02E01.mp4"'
    )
    assert expected_error == str(e.value)


def test_move_season(fs):
    os.makedirs('/tv/Program/Season 1')
    create_program_info(seasons={1: [1]})
    os.chdir('/tv/Program')
    with open('/tv/Program/Season 1/Program - S01E01.mp4', 'w') as f:
        f.write('S01E01')
    Mover('/tv/Program/Season 1', '/tv/Program/Season 2').move()
    assert not os.path.exists('/tv/Program/Season 1')
    assert os.path.exists('/tv/Program/Season 2')
    assert len(os.listdir('/tv/Program/Season 2/')) == 1
    assert os.path.exists('/tv/Program/Season 2/Program - S02E01.mp4')
    pi = ProgramInfo('/tv/Program')
    assert 1 not in pi.seasons
    assert 2 in pi.seasons
    assert len(pi.seasons[2]) == 1
    assert pi.seasons[2][0].episode.number == 1


def test_move_episode_raises_if_moving_between_seasons(fs):
    os.makedirs('/tv/Program/Season 1')
    os.makedirs('/tv/Program/Season 2')
    with open('/tv/Program/Season 1/S01E01.mp4', 'w') as f:
        f.write('S01E01')
    create_program_info(seasons={1: [1]})
    with pytest.raises(RuntimeError) as e:
        Mover(
            '/tv/Program/Season 1/Program - S01E01.mp4',
            '/tv/Program/Season 2/Program - S02E01.mp4'
        ).move()
    assert 'Moving episodes between seasons is not supported' in str(e.value)


def test_move_episode_fails_if_src_episode_not_exists_in_program_info(fs):
    os.makedirs('/tv/Program/Season 1')
    create_program_info(seasons={
        1: []
    })
    with open('/tv/Program/Season 1/S01E01.mp4', 'w') as f:
        f.write('S01E01')
    with pytest.raises(ProgramInfoError) as e:
        Mover(
            '/tv/Program/Season 1/Program - S01E01.mp4',
            '/tv/Program/Season 1/Program - S01E02.mp4'
        ).move()
    assert 'Source episode not found in program info file' in str(e.value)


def test_move_episode_fails_if_dst_episode_exists_in_program_info(fs):
    os.makedirs('/tv/Program/Season 1')
    create_program_info(seasons={
        1: [1, 2]
    })
    with open('/tv/Program/Season 1/S01E01.mp4', 'w') as f:
        f.write('S01E01')
    with pytest.raises(ProgramInfoError) as e:
        Mover(
            '/tv/Program/Season 1/Program - S01E01.mp4',
            '/tv/Program/Season 1/Program - S01E02.mp4'
        ).move()
    assert 'Destination episode found in program info file' in str(e.value)


def test_move_episode_requires_correct_naming_schema(fs):
    os.makedirs('/tv/Program/Season 1/')
    create_program_info(
        seasons={
            1: [1]
        }
    )
    with open('/tv/Program/Season 1/S01E01.mp4', 'w') as f:
        f.write('incorrect name')
    with open('/tv/Program/Season 1/Program 1x1.mp4', 'w') as f:
        f.write('incorrect name')
    with open('/tv/Program/Season 1/NotProgram 1x1.mp4', 'w') as f:
        f.write('incorrect name')
    with open('/tv/Program/Season 1/Program - S01E01.mp4', 'w') as f:
        f.write('correct name')

    with pytest.raises(NamingSchemaError) as e:
        Mover(
            '/tv/Program/Season 1/S01E01.mp4',
            '/tv/Program/Season 1/Program - S01E02.mp4'
        ).move()
    assert 'Episode/season number not determined from' in str(e.value)

    with pytest.raises(NamingSchemaError) as e:
        Mover(
            '/tv/Program/Season 1/Program 1x1.mp4',
            '/tv/Program/Season 1/Program - S01E02.mp4'
        ).move()
    assert 'Episode/season number not determined from' in str(e.value)

    with pytest.raises(NamingSchemaError) as e:
        Mover(
            '/tv/Program/Season 1/NotProgram.mp4',
            '/tv/Program/Season 1/Program - S01E02.mp4'
        ).move()
    assert 'Episode/season number not determined from' in str(e.value)


def test_move_episode(fs):
    os.makedirs('/tv/Program/Season 1/')
    create_program_info(
        seasons={
            1: [1]
        }
    )
    with open('/tv/Program/Season 1/Program - S01E01.mp4', 'w') as f:
        f.write('correct name')
    Mover(
        '/tv/Program/Season 1/Program - S01E01.mp4',
        '/tv/Program/Season 1/Program - S01E02.mp4'
    ).move()
    assert os.path.exists('/tv/Program/Season 1/Program - S01E02.mp4')
    assert not os.path.exists('/tv/Program/Season 1/Program - S01E01.mp4')

    pi = ProgramInfo('/tv/Program')
    assert 1 in pi.seasons
    season = pi.seasons[1]
    assert len(season) == 1
    entry = season[0]
    assert entry.episode.number == 2



import os
from click.testing import CliRunner

from ruv_dl import mv

runner = CliRunner()


def test_mv(fs, mocker):
    os.makedirs('/a/b/c')
    os.chdir('/a/b')
    mover_patch = mocker.patch('ruv_dl.Mover')
    runner.invoke(mv, ['c', 'd'])
    mover_patch.assert_called_once_with('/a/b/c', '/a/b/d')
    mover_patch().move.assert_called_once_with()



#!/usr/bin/env python
import os
import datetime
import logging

from ruv_dl.constants import DATE_FORMAT

logger = logging.getLogger(__name__)


class Episode:
    def __init__(self, data):
        self.data = data or {}
        self.id = self.data.get('id', None)

    @property
    def number(self):
        return self.data.get('number', None)

    @number.setter
    def number(self, number):
        self.data['number'] = number

    def to_dict(self):
        return self.data


class Entry:
    def __init__(
        self, fn, url, date, etag, episode=None
    ):
        self.fn = fn
        self.url = url
        self.date = date
        self.etag = etag
        self.episode = Episode(episode)
        self.target_path = None

    def to_dict(self):
        return {
            'fn': self.fn,
            'url': self.url,
            'date': self.date.strftime(DATE_FORMAT),
            'etag': self.etag,
            'episode': self.episode.to_dict(),
        }

    @classmethod
    def from_dict(cls, data):
        return cls(
            fn=data['fn'],
            url=data['url'],
            date=datetime.datetime.strptime(data['date'], DATE_FORMAT),
            etag=data['etag'],
            episode=data.get('episode'),
        )

    def get_target_basename(self, program, season):
        return (
            f'{program["title"]} - '
            f'S{str(season).zfill(2)}E{str(self.episode.number).zfill(2)}.mp4'
        )

    @classmethod
    def get_season_folder(cls, destination, program, season):
        return os.path.join(
            destination,
            program['title'],
            f'Season {season}',
        )

    def set_target_path(self, path):
        # TODO: Get rid of this maybe
        self.target_path = path

    def exists_on_disk(self):
        if self.target_path is None:
            raise RuntimeError(f'Missing target path for {self.to_dict}')
        return os.path.exists(self.target_path)

    def __hash__(self):
        return hash(self.etag)

    def __eq__(self, other):
        return isinstance(other, Entry) and hash(self) == hash(other)

    def __str__(self):
        return f'{self.fn} - {self.date}'


class EntrySet(set):
    def add(self, item):
        for member in self:
            if member == item:
                # If we have the same item twice, we want the one with a full
                # episode entry, if available, chosen.
                item = self._choose_best_item(item, member)
                self.remove(member)
                break
        super().add(item)

    def sorted(self):
        return sorted(self, key=lambda entry: entry.date)

    @classmethod
    def find_target_number(cls, entries, i):
        '''
            Find a target number for entries[i] wrt entries
        '''
        for k in range(0, len(entries)):
            if entries[k].episode and entries[k].episode.id:
                target = entries[k].episode.number + i - k
                if target < 1:
                    logger.warning(
                        'Found strange episode number. You probably want to '
                        'manually fix this entry: %s',
                        entries[i],
                    )
                    target += len(entries)
                logger.info(
                    'Found related episode in %d for %d: %s, '
                    'target number is %d',
                    k,
                    i,
                    entries[i],
                    target
                )
                return target
        logger.warning(
            'Found no episode number for %s, assigning it 1 '
            '(number of entries: %d)',
            entries[i],
            len(entries),
        )
        return 1

    def _choose_best_item(self, item, member):
        if item.episode and item.episode.id:
            return item
        elif member.episode:
            return member
        return item

    def __getitem__(self, i):
        return self.sorted()[i]



#!/usr/bin/env python
import os
import re
import logging
import shutil

from ruv_dl.programs import ProgramInfo
from ruv_dl.runtime import settings
from ruv_dl.exceptions import NamingSchemaError, ProgramInfoError


SEASON_RE = re.compile(f'^Season (\\d+)$')
logger = logging.getLogger(__name__)


class Mover:
    def __init__(self, src, dst):
        for path in (src, dst):
            if not os.path.isabs(path):
                raise OSError(
                    f'Must pass in absolute path - "{path}" is relative.'
                )
        if os.path.exists(dst):
            raise FileExistsError(f'{dst} exists')
        dirname = os.path.dirname(src)
        while dirname != os.path.dirname(dirname):
            try:
                self.program_info = ProgramInfo(dirname)
                break
            except FileNotFoundError:
                dirname = os.path.dirname(dirname)
        else:
            raise FileNotFoundError(
                f'Could not find program info related to {src}'
            )
        self.src = src
        self.dst = dst

    def move(self):
        if os.path.isdir(self.src):
            self.move_season()
        else:
            self.move_episode()

    def move_episode(self):
        src_season, src_episode = self._get_episode_season_and_number(
            self.program_info.program, self.src
        )
        dst_season, dst_episode = self._get_episode_season_and_number(
            self.program_info.program, self.dst
        )
        if dst_season != src_season:
            raise RuntimeError(
                'Moving episodes between seasons is not supported'
            )
        season_number = self._get_season_number(os.path.dirname(self.src))
        seasons = self.program_info.seasons
        season = seasons[season_number]
        src_entry = None
        for entry in season:
            if entry.episode.number == src_episode:
                if src_entry:
                    raise ProgramInfoError(
                        'Found two entries with same number in program info.'
                    )
                src_entry = entry
            if entry.episode.number == dst_episode:
                raise ProgramInfoError(
                    'Destination episode found in program info file.'
                )
        if not src_entry:
            raise ProgramInfoError(
                'Source episode not found in program info file'
            )
        if settings.dryrun:
            logger.warning(f'Dryrun. Would move {self.src} to {self.dst}.')
        else:
            shutil.move(self.src, self.dst)
            src_entry.episode.number = dst_episode
            self.program_info.seasons = seasons
            self.program_info.write()

    def move_season(self):
        seasons = self.program_info.seasons
        src_season_no = self._get_season_number(self.src)
        dst_season_no = self._get_season_number(self.dst)
        if src_season_no not in seasons:
            raise ProgramInfoError(
                f'Source season number {src_season_no} not found in info file '
                f'{self.program_info.fn}'
            )
        files_in_src = os.listdir(self.src)
        filenames_to_move = []
        for fn in files_in_src:
            src = os.path.join(self.src, fn)
            season, episode = self._get_episode_season_and_number(
                self.program_info.program,
                src
            )
            part_change = f'S{str(season).zfill(2)}'
            assert part_change in src
            dst = src.replace(part_change, f'S{str(dst_season_no).zfill(2)}')
            filenames_to_move.append((src, dst))
        if settings.dryrun:
            logger.warning(f'Dryrun. Would move {self.src} to {self.dst}.')
            for src, dst in filenames_to_move:
                logger.warning(
                    f'Would also move {src} to {dst} prior to moving season'
                )
        else:
            for src, dst in filenames_to_move:
                shutil.move(src, dst)
            shutil.move(self.src, self.dst)
            seasons[dst_season_no] = seasons[src_season_no]
            del seasons[src_season_no]
            self.program_info.seasons = seasons
            self.program_info.write()


    def _get_season_number(self, path):
        name = os.path.basename(path)
        result = SEASON_RE.search(name)
        if not result:
            raise NamingSchemaError(
                f'Season number not determined from {path}'
            )
        return int(result.group(1))

    def _get_episode_re(self, program):
        return re.compile(
            f'Season (\\d+){os.sep}{program["title"]} - S(\\d+)E(\\d+).mp4$'
        )

    def _get_episode_season_and_number(self, program, path):
        result = self._get_episode_re(program).search(path)
        if result:
            season1, season2, episode = result.groups()
            if int(season1) == int(season2):
                return int(season1), int(episode)
            raise NamingSchemaError(
                f'Season number in folder and filename differ in "{path}"'
            )
        raise NamingSchemaError(
            f'Episode/season number not determined from {path}'
        )
